% Hierarchical Tree Structure

ancestor(X,Y)   :- parent(X,Y).                   % Y is an ancestor of X
ancestor(X,Y)   :- parent(X,Z), ancestor(Z,Y).
descendent(X,Y) :- child(X,Y).                    % Y is a descendent of X
descendent(X,Y) :- child(X,Z), decendent(Z,Y).
after(X,Y)      :- next(X,Y).                     % Y is somewhere after X
after(X,Y)      :- next(X,Z), after(Z,Y).           
before(X,Y)     :- previous(X,Y).                 % Y is somewhere before X
before(X,Y)     :- previous(X,Z), before(Z,Y).
sibling(X,Y)    :- parent(Z,X), parent(Z,Y), X \== Y. 

/* Not used in new PL - should do a "patch" to correct for these ...
sentence_parent(    X, Y) :- sentence(X), parent(X, Y).
phrase_parent(      X, Y) :- phrase(  X), parent(X, Y).
word_parent(        X, Y) :- word(    X), parent(X, Y).

% Jump over phrase boundaries
word_next(X, Y) :- word_next_within_phrase(X, Y).
% Too slow with this enabled??????  Turn this on as well as require only FIRST phrase in a triple/double to be specified.
word_next(X, Y) :-
	last_word_in_phrase(     Phrase, X),
	phrase_next(Phrase,  NextPhrase),
	first_word_in_phrase(NextPhrase, Y).

word_previous(X, Y) :- word_previous_within_phrase(X, Y).
word_previous(X, Y) :-
	first_word_in_phrase(           Phrase, X),
	phrase_previous(Phrase, PreviousPhrase),
	last_word_in_phrase(    PreviousPhrase, Y).

word_next_within_phrase(    X, Y) :- word(X), next(    X, Y).
word_previous_within_phrase(X, Y) :- word(X), previous(X, Y).

abstract_child(     X, Y) :- abstract(X), child(     X, Y).
sentence_child(     X, Y) :- sentence(X), child(     X, Y).
phrase_child(       X, Y) :- phrase(  X), child(     X, Y).

phrase_next(        X, Y) :- phrase(X), next( X, Y).
phrase_previous(    X, Y) :- phrase(X), previous(X, Y).

*/
parent(X, Y) :- sentence_parent(X, Y).
parent(X, Y) :- phrase_parent(  X, Y).
parent(X, Y) :- word_parent(    X, Y).

child( X, Y) :- abstract_child( X, Y).
child( X, Y) :- sentence_child( X, Y).
child( X, Y) :- phrase_child(   X, Y).

next(  X, Y) :- sentence_next(  X, Y).
next(  X, Y) :- phrase_next(    X, Y).
next(  X, Y) :- word_next(      X, Y).

previous(X, Y) :- sentence_previous(X, Y).
previous(X, Y) :- phrase_previous(  X, Y).
previous(X, Y) :- word_previous(    X, Y).



phrase_ancestor(    X, Y) :- phrase(X), ancestor(X, Y).
word_ancestor(      X, Y) :- word(  X), ancestor(X, Y).

sentence_descendent(X, Y) :- sentence(X), descendent(X, Y).
phrase_descendent(  X, Y) :- phrase(  X), descendent(X, Y).

phrase_after(       X, Y) :- phrase(X), after(X, Y).
word_after_within_phrase(X, Y) :- word(X), word_parent(X, Z), word_parent(Y, Z), after(X, Y).

phrase_before(      X, Y) :- phrase(X), before(  X, Y).
word_before_within_phrase(X, Y)  :- word(X), word_parent(X, Z), word_parent(Y, Z), before(X, Y).

phrase_sibling(     X, Y) :- phrase(X), sibling(X, Y).
word_sibling_within_phrase(X, Y) :- word(X), word_parent(X, Z), word_parent(Y, Z), sibling(X, Y).



:- modeb(1,             sentence_parent(    +phrase,   -abstract)).
:- modeb(5, sentence_child(     +sentence, -phrase)). % Turn these off to constrain search?  Ie, can only each other phrases from other arguments in the head.
:- modeb(5, sentence_descendent(+sentence, -phrase)).
:- modeb(9, sentence_descendent(+sentence, -word)).

:- modeb(3, phrase_ancestor(    +phrase,   -sentence)).
:- modeb(9, phrase_descendent(  +phrase,   -word)).
:- modeb(9, phrase_child(       +phrase,   -word)).
:- modeb(1,             phrase_parent(      +phrase,   -sentence)).
:- modeb(1,             phrase_next(        +phrase,   -phrase)).
:- modeb(1,             phrase_previous(    +phrase,   -phrase)).
:- modeb(5, phrase_sibling(     +phrase,   -phrase)).
:- modeb(3, phrase_before(      +phrase,   -phrase)).
:- modeb(3, phrase_after(       +phrase,   -phrase)).

:- modeb(1,             word_ancestor(      +word,     -phrase)).
:- modeb(1,             word_ancestor(      +word,     -sentence)).
:- modeb(1,             word_parent(        +word,     -phrase)).
:- modeb(1,             word_next_within_phrase(    +word, -word)).
:- modeb(1,             word_next(                  +word, -word)).
:- modeb(1,             word_previous_within_phrase(+word, -word)).
:- modeb(1,             word_previous(              +word, -word)).
:- modeb(7, word_before(                +word, -word)).
:- modeb(7, word_after(                 +word, -word)).
:- modeb(7, word_sibling_within_phrase( +word, -word)).
:- modeb(3, word_before_within_phrase(  +word, -word)).
:- modeb(3, word_after_within_phrase(   +word, -word)).

different_phrases(X, Y) :- phrase(X), phrase(Y), X \== Y.
different_words(X, Y)   :- word(  X), word(  Y), X \== Y.  % Too many of these in the bottom clause?  So probably shouldn't use.
:- symmetric(different_phrases/2). % Does aleph make use of this?  Or is this just some untested code?
:- symmetric(different_words/2).
:- modeb(1,             different_phrases(+phrase, +phrase)). % Both args must be bound.
%:-modeb(1,             different_words(+word, +word)).


% Assertions that allow mapping from our atom names to the actual text strings.
% (We use these atoms since the same phrase or word can appear multiple times.)

% So that specific words can be added in one search step, define some inference rules.
phrase_contains_specific_word(            Phrase, Word, SpecificWord)
	:- phrase_child(Phrase, Word), word_ID_to_string(Word, SpecificWord).

phrase_contains_specific_word_pair(       Phrase, Word1, Word2,        SpecificWord1, SpecificWord2)
	:- phrase_child(Phrase, Word1), word_ID_to_string(Word1, SpecificWord1), word_next_within_phrase(Word1, Word2), % Stay in phrase.
	   phrase_child(Phrase, Word2), word_ID_to_string(Word2, SpecificWord2).
phrase_contains_specific_word_triple(     Phrase, Word1, Word2, Word3, SpecificWord1, SpecificWord2, SpecificWord3)
	:- phrase_child(Phrase, Word1), word_ID_to_string(Word1, SpecificWord1), word_next_within_phrase(Word1, Word2), % Stay in phrase.
	   phrase_child(Phrase, Word2), word_ID_to_string_extra(  Word2, SpecificWord2), word_next_within_phrase(Word2, Word3),
	   phrase_child(Phrase, Word3), word_ID_to_string(Word3, SpecificWord3).

phrase_contains_POS(                      Phrase, Word, Type)
	:- phrase_child(Phrase, Word), wordPOS(          Word, Type).
phrase_contains_POS_pair(       Phrase, Word1, Word2,        Type1, Type2)
	:- phrase_child(Phrase, Word1), wordPOS(Word1, Type1), word_next_within_phrase(Word1, Word2), % Stay in phrase.
	   phrase_child(Phrase, Word2), wordPOS(Word2, Type2).
phrase_contains_POS_triple(     Phrase, Word1, Word2, Word3, Type1, Type2, Type3)
	:- phrase_child(Phrase, Word1), wordPOS(      Word1, Type1), word_next_within_phrase(Word1, Word2), % Stay in phrase.
	   phrase_child(Phrase, Word2), wordPOS_extra(Word2, Type2), word_next_within_phrase(Word2, Word3),
	   phrase_child(Phrase, Word3), wordPOS(      Word3, Type3).

sentence_contains_specific_word(Sentence, Phrase, Word, SpecificWord) % Seems to overfit ...
	:- sentence_child(Sentence, Phrase), phrase_contains_specific_word(Phrase, Word, SpecificWord).
sentence_contains_specific_phrase(Sentence, Phrase, SpecificPhrase) 
	:- sentence_child(Sentence, Phrase), phrase_ID_to_string(Phrase, SpecificPhrase).
sentence_contains_specific_word_pair(Sentence, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2)
	:- sentence_child(Sentence, Phrase1), phrase_child(Phrase1, Word1), word_ID_to_string(Word1, SpecificWord1), word_next(Word1, Word2),
	   word_ID_to_string(Word2, SpecificWord2), phrase_child(Phrase2, Word2).
sentence_contains_POS_pair(Sentence, Phrase1, Phrase2, Word1, Word2,        Type1, Type2)
	:- sentence_child(Sentence, Phrase1), phrase_child(Phrase1, Word1), wordPOS(Word1, Type1), word_next(Word1, Word2),
	   wordPOS(Word2, Type2), phrase_child(Phrase2, Word2).

% Look for, say: the word 'gene' followed by a verb.
sentence_contains_specific_word_POS_pair(Sentence, Phrase1, Phrase2, Word1, Word2,        SpecificWord1, Type2)
	:- sentence_child(Sentence, Phrase1), phrase_child(Phrase1, Word1), word_ID_to_string(Word1, SpecificWord1), word_next(Word1, Word2),
	   sentence_child(Phrase2, Word2), wordPOS(Word2, Type2).
% Look for, say: an alphanumeric followed by the word 'disease'.
sentence_contains_specific_POS_word_pair(Sentence, Phrase1, Phrase2, Word1, Word2,        Type1, SpecificWord2)
	:- sentence_child(Sentence, Phrase1), phrase_child(Phrase1, Word1), wordPOS(Word1, Type1), word_next(Word1, Word2),
	   sentence_child(Phrase2, Word2), word_ID_to_string(Word2, SpecificWord2).
	   
% Look for triples of words, allowing for a wild-card in the middle.
sentence_contains_specific_word_triple(Sentence, Phrase1, Phrase2, Phrase3, Word1, Word2, Word3, SpecificWord1, SpecificWord2, SpecificWord3)
	:- sentence_child(Sentence, Phrase1), phrase_child(Phrase1, Word1),
				     word_ID_to_string(Word1, SpecificWord1),                               word_next(Word1, Word2),
	                             word_ID_to_string_extra(  Word2, SpecificWord2), phrase_child(Phrase2, Word2), word_next(Word2, Word3), % Allow "dont care" in the middle (no need for the outer words).
	                             word_ID_to_string(Word3, SpecificWord3), phrase_child(Phrase3, Word3).
% Look for triples of word types, allowing for a wild-card in the middle.
sentence_contains_POS_triple(Sentence, Phrase1, Phrase2, Phrase3, Word1, Word2, Word3, Type1, Type2, Type3)
	:- sentence_child(Sentence, Phrase1), phrase_child(Phrase1, Word1),
				     wordPOS(      Word1, Type1),                             word_next(Word1, Word2),
				     wordPOS_extra(Word2, Type2), phrase_child(Phrase2, Word2), word_next(Word2, Word3), % Allow "dont care" in the middle (no need for the outer words).
	                             wordPOS(      Word3, Type3), phrase_child(Phrase3, Word3).

:- modeb(1,             word_ID_to_string(    +word,     #string)). % Access the actual word.
:- modeb(1,             phrase_ID_to_string(  +phrase,   #string)). % Access the actual phrase.
:- modeb(1,             sentence_ID_to_string(+sentence, #string)). % Access the actual sentence. (No need to access the actual sentence since hopefully it is unique or nearly so.)

:- modeb(5, phrase_contains_specific_word(           +phrase,   -word,                   #string)).
:- modeb(3, phrase_contains_specific_word_pair(      +phrase,   -word,   -word,          #string, #string)).
:- modeb(3, phrase_contains_specific_word_triple(    +phrase,   -word,   -word,   -word, #string, #string, #string)).
:- modeb(3, phrase_contains_POS(                     +phrase,   -word,                   #pos)).
:- modeb(3, phrase_contains_POS_pair(                +phrase,   -word,   -word,          #pos,    #pos)).
:- modeb(3, phrase_contains_POS_triple(              +phrase,   -word,   -word,   -word, #pos,    #pos, #pos)).

:- modeb(3, sentence_contains_specific_word(         +sentence, -phrase, -word,               #string)).
:- modeb(3, sentence_contains_specific_phrase(       +sentence, -phrase,                      #string)).

%%%%%
% Require the first phrase to be bound (to constrain search).
% Use 3 here since this is PER PHRASE (since 'phrase' is a bound arg).
:- modeb(3, sentence_contains_specific_word_pair(    +sentence, +phrase, -phrase,          -word, -word,        #string, #string)).
:- modeb(3, sentence_contains_specific_word_triple(  +sentence, +phrase, -phrase, -phrase, -word, -word, -word, #string, #string, #string)).
:- modeb(3, sentence_contains_POS_pair(              +sentence, +phrase, -phrase,          -word, -word,        #pos,    #pos)).
:- modeb(3, sentence_contains_POS_triple(            +sentence, +phrase, -phrase, -phrase, -word, -word, -word, #pos,    #pos,    #pos)).

:- modeb(3, sentence_contains_specific_word_POS_pair(+sentence, +phrase, -phrase,          -word, -word,        #string, #pos)).
:- modeb(3, sentence_contains_specific_POS_word_pair(+sentence, +phrase, -phrase,          -word, -word,        #pos,    #string)).
%%%%%

word_before_both_target_phrases(ExampleID, Dataset, Fold, Phrase, Word, SpecificWord) :-
	before_both_target_phrases(ExampleID, Dataset, Fold, Phrase),
	phrase_child(Phrase, Word), word_ID_to_string(Word, SpecificWord).

word_in_between_target_phrases( ExampleID, Dataset, Fold, Phrase, Word, SpecificWord) :-
	in_between_target_phrases( ExampleID, Dataset, Fold, Phrase),
	phrase_child(Phrase, Word), word_ID_to_string(Word, SpecificWord).

word_after_both_target_phrases( ExampleID, Dataset, Fold, Phrase, Word, SpecificWord) :-
	after_both_target_phrases( ExampleID, Dataset, Fold, Phrase),
	phrase_child(Phrase, Word), word_ID_to_string(Word, SpecificWord).


% Actually, the second word can be inside a target phrase, but that is ok.
% (Too inefficient to also allow the FIRST word to be in a target phrase.  Besides, a phrase inside a target phrase can run long.)
word_pair_in_between_target_phrases(ExampleID, Dataset, Fold, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2) :-
	in_between_target_phrases(  ExampleID, Dataset, Fold, Phrase1),
	sentence_contains_specific_word_pair(    _,           Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2).
pos_pair_in_between_target_phrases( ExampleID, Dataset, Fold, Phrase1, Phrase2, Word1, Word2, Type1,         Type2) :-
	in_between_target_phrases(  ExampleID, Dataset, Fold, Phrase1),
	sentence_contains_POS_pair(              _,           Phrase1, Phrase2, Word1, Word2, Type1,         Type2).
word_pos_in_between_target_phrases( ExampleID, Dataset, Fold, Phrase1, Phrase2, Word1, Word2, SpecificWord1, Type2) :-
	in_between_target_phrases(  ExampleID, Dataset, Fold, Phrase1),
	sentence_contains_specific_word_POS_pair(_,           Phrase1, Phrase2, Word1, Word2, SpecificWord1, Type2).
pos_word_in_between_target_phrases( ExampleID, Dataset, Fold, Phrase1, Phrase2, Word1, Word2, Type1,         SpecificWord2) :-
	in_between_target_phrases(  ExampleID, Dataset, Fold, Phrase1),
	sentence_contains_specific_POS_word_pair(_,           Phrase1, Phrase2, Word1, Word2, Type1,         SpecificWord2).


word_prev_target_arg1(ExampleID, Dataset, Fold, Phrase, Word, SpecificWord) :-
	phrase_prev_target_arg1(ExampleID, Dataset, Fold, Phrase),
	phrase_child(Phrase, Word), word_ID_to_string(Word, SpecificWord).
word_prev_target_arg2(ExampleID, Dataset, Fold, Phrase, Word, SpecificWord) :-
	phrase_prev_target_arg1(ExampleID, Dataset, Fold, Phrase),
	phrase_child(Phrase, Word), word_ID_to_string(Word, SpecificWord).
word_next_target_arg1(ExampleID, Dataset, Fold, Phrase, Word, SpecificWord) :-
	phrase_next_target_arg1(ExampleID, Dataset, Fold, Phrase),
	phrase_child(Phrase, Word), word_ID_to_string(Word, SpecificWord).
word_next_target_arg2(ExampleID, Dataset, Fold, Phrase, Word, SpecificWord) :-
	phrase_next_target_arg1(ExampleID, Dataset, Fold, Phrase),
	phrase_child(Phrase, Word), word_ID_to_string(Word, SpecificWord).

word_pair_prev_target_arg1(ExampleID, Dataset, Fold, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2) :-
	phrase_prev_target_arg1(ExampleID, Dataset, Fold, Phrase1),
	sentence_contains_specific_word_pair(    _, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2).
word_pair_prev_target_arg2(ExampleID, Dataset, Fold, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2) :-
	phrase_prev_target_arg2(ExampleID, Dataset, Fold, Phrase1),
	sentence_contains_specific_word_pair(    _, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2).

word_pair_next_target_arg1(ExampleID, Dataset, Fold, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2) :-
	phrase_next_target_arg1(ExampleID, Dataset, Fold, Phrase1),
	sentence_contains_specific_word_pair(    _, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2).
word_pair_next_target_arg2(ExampleID, Dataset, Fold, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2) :-
	phrase_next_target_arg2(ExampleID, Dataset, Fold, Phrase1),
	sentence_contains_specific_word_pair(    _, Phrase1, Phrase2, Word1, Word2, SpecificWord1, SpecificWord2).


few_phrases_before_target_args(    ExampleID, Dataset, Fold) :-
	phrases_before_target_args(ExampleID, Dataset, Fold, Count), Count > 0, Count < 4.
several_phrases_before_target_args(ExampleID, Dataset, Fold) :-
	phrases_before_target_args(ExampleID, Dataset, Fold, Count), Count > 3, Count < 8.
many_phrases_before_target_args(   ExampleID, Dataset, Fold) :-
	phrases_before_target_args(ExampleID, Dataset, Fold, Count), Count > 7.

few_words_before_target_args(    ExampleID, Dataset, Fold) :-
	words_before_target_args(ExampleID, Dataset, Fold, Count), Count > 0, Count < 5.
several_words_before_target_args(ExampleID, Dataset, Fold) :-
	words_before_target_args(ExampleID, Dataset, Fold, Count), Count > 4, Count < 10.
many_words_before_target_args(   ExampleID, Dataset, Fold) :-
	words_before_target_args(ExampleID, Dataset, Fold, Count), Count > 9.

:- modeb(1, few_phrases_before_target_args(    +example, +dataset, #fold)).
:- modeb(1, several_phrases_before_target_args(+example, +dataset, #fold)).
:- modeb(1, many_phrases_before_target_args(   +example, +dataset, #fold)).
:- modeb(1, few_words_before_target_args(      +example, +dataset, #fold)).
:- modeb(1, several_words_before_target_args(  +example, +dataset, #fold)).
:- modeb(1, many_words_before_target_args(     +example, +dataset, #fold)).

few_phrases_between_target_args(    ExampleID, Dataset, Fold) :-
	phrases_between_target_args(ExampleID, Dataset, Fold, Count), Count > 0, Count < 4.
several_phrases_between_target_args(ExampleID, Dataset, Fold) :-
	phrases_between_target_args(ExampleID, Dataset, Fold, Count), Count > 3, Count < 8.
many_phrases_between_target_args(   ExampleID, Dataset, Fold) :-
	phrases_between_target_args(ExampleID, Dataset, Fold, Count), Count > 7.

few_words_between_target_args(    ExampleID, Dataset, Fold) :-
	words_between_target_args(ExampleID, Dataset, Fold, Count), Count > 0, Count < 5.
several_words_between_target_args(ExampleID, Dataset, Fold) :-
	words_between_target_args(ExampleID, Dataset, Fold, Count), Count > 4, Count < 10.
many_words_between_target_args(   ExampleID, Dataset, Fold) :-
	words_between_target_args(ExampleID, Dataset, Fold, Count), Count > 9.

:- modeb(1, few_phrases_between_target_args(    +example, +dataset, #fold)).
:- modeb(1, several_phrases_between_target_args(+example, +dataset, #fold)).
:- modeb(1, many_phrases_between_target_args(   +example, +dataset, #fold)).
:- modeb(1, few_words_between_target_args(      +example, +dataset, #fold)).
:- modeb(1, several_words_between_target_args(  +example, +dataset, #fold)).
:- modeb(1, many_words_between_target_args(     +example, +dataset, #fold)).

few_phrases_after_target_args(    ExampleID, Dataset, Fold) :-
	phrases_after_target_args(ExampleID, Dataset, Fold, Count), Count > 0, Count < 4.
several_phrases_after_target_args(ExampleID, Dataset, Fold) :-
	phrases_after_target_args(ExampleID, Dataset, Fold, Count), Count > 3, Count < 8.
many_phrases_after_target_args(   ExampleID, Dataset, Fold) :-
	phrases_after_target_args(ExampleID, Dataset, Fold, Count), Count > 7.

few_words_after_target_args(    ExampleID, Dataset, Fold) :-
	words_after_target_args(ExampleID, Dataset, Fold, Count), Count > 0, Count < 5.
several_words_after_target_args(ExampleID, Dataset, Fold) :-
	words_after_target_args(ExampleID, Dataset, Fold, Count), Count > 4, Count < 10.
many_words_after_target_args(   ExampleID, Dataset, Fold) :-
	words_after_target_args(ExampleID, Dataset, Fold, Count), Count > 9.

:- modeb(1, few_phrases_after_target_args(    +example, +dataset, #fold)).
:- modeb(1, several_phrases_after_target_args(+example, +dataset, #fold)).
:- modeb(1, many_phrases_after_target_args(   +example, +dataset, #fold)).
:- modeb(1, few_words_after_target_args(      +example, +dataset, #fold)).
:- modeb(1, several_words_after_target_args(  +example, +dataset, #fold)).
:- modeb(1, many_words_after_target_args(     +example, +dataset, #fold)).

% Capture words BEFORE  the two phrases in the head.
:- modeb(3, before_both_target_phrases(     +example, +dataset, #fold, -phrase)).
:- modeb(7, word_before_both_target_phrases(+example, +dataset, #fold, -phrase, -word, #string)).

% Capture words BETWEEN the two phrases in the head.
:- modeb(5, in_between_target_phrases(      +example, +dataset, #fold, -phrase)).
:- modeb(9, word_in_between_target_phrases( +example, +dataset, #fold, -phrase, -word, #string)).

% Capture words AFTER   the two phrases in the head.
:- modeb(3, after_both_target_phrases(      +example, +dataset, #fold, -phrase)).
:- modeb(7, word_after_both_target_phrases( +example, +dataset, #fold, -phrase, -word, #string)).


:- modeb(1, target_arg1_before_target_arg2(+example, +dataset, #fold)).
:- modeb(1, target_arg2_before_target_arg1(+example, +dataset, #fold)).

% Consider more word-pairs in between the args than outside.
:- modeb(5, word_pair_in_between_target_phrases(+example, +dataset, #fold, -phrase, -phrase, -word, -word, #string, #string)).
:- modeb(5, pos_pair_in_between_target_phrases( +example, +dataset, #fold, -phrase, -phrase, -word, -word, #pos,    #pos)).
:- modeb(5, word_pos_in_between_target_phrases( +example, +dataset, #fold, -phrase, -phrase, -word, -word, #string, #pos)).
:- modeb(5, pos_word_in_between_target_phrases( +example, +dataset, #fold, -phrase, -phrase, -word, -word, #pos,    #string)).

:- modeb(7, word_prev_target_arg1(              +example, +dataset, #fold, -phrase, -word, #string)).
:- modeb(7, word_prev_target_arg2(              +example, +dataset, #fold, -phrase, -word, #string)).
:- modeb(7, word_next_target_arg1(              +example, +dataset, #fold, -phrase, -word, #string)).
:- modeb(7, word_next_target_arg2(              +example, +dataset, #fold, -phrase, -word, #string)).

:- modeb(3, word_pair_prev_target_arg1(         +example, +dataset, #fold, -phrase, -phrase, -word, -word, #string, #string)).
:- modeb(3, word_pair_prev_target_arg2(         +example, +dataset, #fold, -phrase, -phrase, -word, -word, #string, #string)).
:- modeb(3, word_pair_next_target_arg1(         +example, +dataset, #fold, -phrase, -phrase, -word, -word, #string, #string)).
:- modeb(3, word_pair_next_target_arg2(         +example, +dataset, #fold, -phrase, -phrase, -word, -word, #string, #string)).


set_sentence_structure_determinations(Head) :- 

   uw_setting(use_target_stuff_trainset, UseTargetArgsStuff),
   (UseTargetArgsStuff ->
    determination(Head, target_arg1_before_target_arg2/3),
    determination(Head, target_arg2_before_target_arg1/3),

    determination(Head, before_both_target_phrases/4),
    determination(Head, in_between_target_phrases/4),
    determination(Head, after_both_target_phrases/4),

    determination(Head, few_phrases_before_target_args/3),
    determination(Head, several_phrases_before_target_args/3),
    determination(Head, many_phrases_before_target_args/3),
    determination(Head, few_words_before_target_args/3),
    determination(Head, several_words_before_target_args/3),
    determination(Head, many_words_before_target_args/3),

    determination(Head, few_phrases_between_target_args/3),
    determination(Head, several_phrases_between_target_args/3),
    determination(Head, many_phrases_between_target_args/3),
    determination(Head, few_words_between_target_args/3),
    determination(Head, several_words_between_target_args/3),
    determination(Head, many_words_between_target_args/3),

    determination(Head, few_phrases_after_target_args/3),
    determination(Head, several_phrases_after_target_args/3),
    determination(Head, many_phrases_after_target_args/3),
    determination(Head, few_words_after_target_args/3),
    determination(Head, several_words_after_target_args/3),
    determination(Head, many_words_after_target_args/3),

    determination(Head, word_before_both_target_phrases/6),
    determination(Head, word_in_between_target_phrases/6),
    determination(Head, word_after_both_target_phrases/6),

    determination(Head, word_pair_in_between_target_phrases/9),
    determination(Head, pos_pair_in_between_target_phrases/9),
    determination(Head, word_pos_in_between_target_phrases/9),
    determination(Head, pos_word_in_between_target_phrases/9),

    determination(Head, word_prev_target_arg1/6),
    determination(Head, word_prev_target_arg2/6),
    determination(Head, word_next_target_arg1/6),
    determination(Head, word_next_target_arg2/6),

    determination(Head, word_pair_prev_target_arg1/9), % to do: extend to pos and word+pos (and vice-versa)
    determination(Head, word_pair_prev_target_arg2/9),
    determination(Head, word_pair_next_target_arg1/9),
    determination(Head, word_pair_next_target_arg2/9),
    true ; true),

%  determination(Head, abstract_child/2),

%  determination(Head, sentence_descendent/2), % Descendents are not relevant in a flat parser (I - JWS - assume flat means "one-level").
%  determination(Head, sentence_child/2),

%  determination(Head, phrase_ancestor/2),
%  determination(Head, phrase_descendent/2),
   determination(Head, phrase_child/2),
%  determination(Head, phrase_parent/2), % Turned off for now since not needed due to flattened parser.
   determination(Head, phrase_next/2),
   determination(Head, phrase_previous/2),
   determination(Head, phrase_before/2),
   determination(Head, phrase_after/2),
%  determination(Head, phrase_sibling/2),

%  determination(Head, word_ancestor/2),
%  determination(Head, word_parent/2), % Turned off for now since not needed due to flattened parser.
   determination(Head, word_next_within_phrase/2),
   determination(Head, word_next/2),
   determination(Head, word_previous_within_phrase/2),
   determination(Head, word_previous/2),
   determination(Head, word_before_within_phrase/2),
   determination(Head, word_after_within_phrase/2),
   determination(Head, word_sibling_within_phrase/2),
%  determination(Head, word_before/2), % Not defined yet.
%  determination(Head, word_after/2),  % Not defined yet.

% Allow aleph to REQUIRE that two variables be different.
   determination(Head, different_phrases/2),
%  determination(Head, different_words/2),

   determination(Head, phrase_contains_specific_word/3),
   determination(Head, phrase_contains_specific_word_pair/5),
   determination(Head, phrase_contains_specific_word_triple/7),
   determination(Head, phrase_contains_POS/3),
   determination(Head, phrase_contains_POS_pair/5),
   determination(Head, phrase_contains_POS_triple/7),

   determination(Head, sentence_contains_specific_word_POS_pair/7), % TO DO: create phrase'd versions of these?
   determination(Head, sentence_contains_specific_POS_word_pair/7),

   determination(Head, sentence_contains_specific_word/4),
   determination(Head, sentence_contains_specific_phrase/3),
   determination(Head, sentence_contains_specific_word_pair/7),
   determination(Head, sentence_contains_specific_word_triple/10),
   determination(Head, sentence_contains_POS_pair/7),
   determination(Head, sentence_contains_POS_triple/10),

   determination(Head, first_word_in_phrase/2),
   determination(Head, last_word_in_phrase/2),
   determination(Head, first_phrase_in_sentence/2),
   determination(Head, last_phrase_in_sentence/2),

   true.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% 
%   Some utilities
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


word_in_fold(Word, Dataset, Fold) :-
	word(Word),
	word_parent(Word, Phrase),
	phrase_in_fold(Phrase, Dataset, Fold).
phrase_in_fold(Phrase, Dataset, Fold) :-
	phrase(Phrase),
	phrase_parent(Phrase, Sentence),
	sentence_in_fold(Sentence, Dataset, Fold).
sentence_in_fold(Sentence, Dataset, Fold) :-
	sentence(Sentence),
	sentence_parent( Sentence, Abstract),
	abstract_in_fold(Abstract, Dataset, Fold).
abstract_in_fold(Abstract, Dataset, Fold) :-
	abstract(Abstract),
	assigned_subfold_for_abstract(Abstract, Subfold),
	assigned_fold(Subfold, Dataset, Fold).

word_in_subfold(Word, Subfold) :-
	word(Word),
	word_parent(Word, Phrase),
	phrase_in_subfold(Phrase, Subfold).
phrase_in_subfold(Phrase, Subfold) :-
	phrase(Phrase),
	phrase_parent(Phrase, Sentence),
	sentence_in_subfold(Sentence, Subfold).
sentence_in_subfold(Sentence, Subfold) :-
	sentence(Sentence),
	sentence_parent(Sentence, Abstract),
	abstract_in_subfold(Abstract, Subfold).
abstract_in_subfold(Abstract, Subfold) :-
	abstract(Abstract),
	assigned_subfold_for_abstract(Abstract, Subfold).


contained_in_sentence(X, Sentence) :-   % X can be a word or phrase.
	word(X), !,
	word_parent(X, Phrase),
	contained_in_sentence(Phrase, Sentence).	
contained_in_sentence(X, Sentence) :-
	phrase(X), !,
	phrase_parent(X, Sentence).
contained_in_sentence(X, _) :- 
	told,
	uw_format("~p must be a word or phrase to be used in contained_in_sentence(X, Sentence).~n", [X]),
	break.

contained_in_abstract(X, Abstract) :-   % X can be a word, phrase, or sentence
	word(X), !,
	word_parent(X, Phrase),
	contained_in_abstract(Phrase, Abstract).	
contained_in_abstract(X, Abstract) :-
	phrase(X), !,
	phrase_parent(X, Sentence),
	contained_in_abstract(Sentence, Abstract).
contained_in_abstract(X, Abstract) :-
	sentence(X), !,
	sentence_parent(X, Abstract).
contained_in_abstract(X, _) :- 
	told,
	uw_format("~p must be a word, phrase, or sentence to be used in contained_in_abstract(X, Abstract).~n", [X]),
	break.

%%%%%%%%%%%%%%%%%%%%%%%%%
% First arg is INPUT, others can be OUTPUT.  These do assume Sentence, Phrase, and/or Word
% are already defined (since parent(X,Y) is called in some cases).

% (+Sentence, -SentenceNumber, -Abstract)
break_down_sentence_token(Sentence, SentenceNumber, Abstract) :-
	nonvar(Sentence),
	uw_confirm(sentence(Sentence), 'Must provide a sentence to break_down_sentence_token.'),
	name('_', [UnderscoreChar]),	
	name(Sentence, CharsInS),
	uw_collect_chars_up_to(UnderscoreChar, CharsInS, PrefixCharsInS),
	name(Abstract, PrefixCharsInS),
	uw_reverse(CharsInS, ReverseCharsInS),
	name('n', [N_Char]),	% Grab DIGITS at end (eg, "sen17").
	uw_collect_chars_up_to(N_Char, ReverseCharsInS, FinalCharsInS_reversed),
	uw_reverse(FinalCharsInS_reversed, FinalCharsInS),
	name(SentenceNumber, FinalCharsInS).
break_down_sentence_token(Sentence, _, _) :-
	told, uw_format("Error in break_down_sentence_token: ~p~n", [Sentence]), break.

% (+Phrase, -PhraseNumber, -Sentence, -SentenceNumber, -Abstract)
break_down_phrase_token(Phrase, PhraseNumber, Sentence, SentenceNumber, Abstract) :-
	nonvar(Phrase),
	uw_confirm(phrase(Phrase), 'Must provide a phrase to break_down_phrase_token.'),
	name(Phrase, CharsInP),
	uw_reverse(CharsInP, ReverseCharsInP),
	name('h', [H_Char]),	% Grab DIGITS at end (eg, "ph23").
	uw_collect_chars_up_to(H_Char, ReverseCharsInP, FinalCharsInP_reversed),
	uw_reverse(FinalCharsInP_reversed, FinalCharsInP),
	name(PhraseNumber, FinalCharsInP),
	parent(Phrase, Sentence), % Assumes we have this mapping ...  (otherwise need a more general rewrite).
	break_down_sentence_token(Sentence, SentenceNumber, Abstract), !.
break_down_phrase_token(Phrase, _, _, _, _) :-
	told, uw_format("Error in break_down_phrase_token: ~p~n", [Phrase]), break.
		
% (+Word, -WordNumber, -Phrase, -PhraseNumber, -Sentence, -SentenceNumber, -Abstract)
break_down_word_token(Word, WordNumber, Phrase, PhraseNumber, Sentence, SentenceNumber, Abstract) :-
	nonvar(Word),
	uw_confirm(word(Word), 'Must provide a word to break_down_word_token.'),
	name(Word, CharsInW),
	uw_reverse(CharsInW, ReverseCharsInW),
	name('w', [W_Char]),	% Grab DIGITS at end (eg, "w52").
	uw_collect_chars_up_to(W_Char, ReverseCharsInW, FinalCharsInW_reversed),
	uw_reverse(FinalCharsInW_reversed, FinalCharsInW),
	name(WordNumber, FinalCharsInW),
	parent(Word, Phrase), % Assumes we have this mapping ...  (otherwise need a more general rewrite).
	break_down_phrase_token(Phrase, PhraseNumber, Sentence, SentenceNumber, Abstract), !.
break_down_word_token(Word, _, _, _, _, _, _) :-
	told, uw_format("Error in break_down_word_token: ~p~n", [Word]), break.
